/* Copyright 2022, Michele Santullo
 * This file is part of duckticker.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with duckticker.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "bitcoinity_reader.hpp"
#include <string_view>
#include <cassert>

namespace duck {
namespace {
	constexpr std::string_view BitcoinityHostname = "bitcoinity.org";
	constexpr std::string_view BitcoinityPort = "80";
	constexpr std::string_view BitcoinityWSPath = "/webs_bridge/websocket";
	constexpr std::string_view JsonKeyPayload = BitcoinityPayload::JsonKeyPayload;
	constexpr std::string_view JsonKeyRef = "ref";
	constexpr std::string_view JsonKeyEvent = "event";
	constexpr std::string_view JsonKeyTopic = "topic";

	std::string force_to_string(const std::optional<std::string_view>& data) {
		if (data)
			return std::string{*data};
		else
			return {};
	}

	BitcoinityMessage to_bitcoinity_message (const std::string& message) {
		if (message.empty())
			throw BitcoinityError("Unable to make a BitcoinityMessage from an empty message");

		simdjson::ondemand::parser parser;
		simdjson::padded_string body = message;
		simdjson::ondemand::document doc = parser.iterate(body);

		auto event_result = doc[JsonKeyEvent];
		if (event_result.error())
			throw BitcoinityMissingKeyError({std::string(JsonKeyEvent)}, message);
		std::string_view event = event_result;

		auto payload_result = doc[JsonKeyPayload];
		if (payload_result.error())
			throw BitcoinityMissingKeyError({std::string(JsonKeyPayload)}, message);

		auto ref_result = doc[JsonKeyRef];
		if (ref_result.error())
			throw BitcoinityMissingKeyError({std::string(JsonKeyRef)}, message);
		std::string_view ref = (ref_result.is_null() ? std::string_view{} : ref_result.value());

		auto topic_result = doc[JsonKeyTopic];
		if (topic_result.error())
			throw BitcoinityMissingKeyError({std::string(JsonKeyTopic)}, message);
		std::string_view topic = topic_result;

		return BitcoinityMessage{
			BitcoinityPayload{std::move(body), std::move(parser)},
			std::move(ref),
			std::move(event),
			std::move(topic)
		};
	}

	void init_bitcoinity (WebsocketReader& ws, std::string_view exchange, std::string_view currency) {
		//thanks to blueagain from #javascript libera.chat for helping figuring
		//out the websocket address and the correct sequence of messages to
		//initiate the communication properly.
		//https://github.com/terrablue
		//1) {"topic":"all","event":"phx_join","payload":{},"ref":"1"}
		//2) {"topic":"webs:markets","event":"phx_join","payload":{},"ref":"2"}
		//3) {"topic":"webs:markets_bitfinex_USD","event":"phx_join","payload":{},"ref":"3"}
		//See docs/bitcoinity/sample_communication.md for a real world example

		{
			ws.write(R"({"topic":"all","event":"phx_join","payload":{},"ref":"1"})");
			auto resp = to_bitcoinity_message(ws.read());
			if (resp.event != "phx_reply" or resp.payload.get_string_or_empty("status") != "ok" or not resp.ref or *resp.ref != "1")
				throw BitcoinityError("Server responded to message 1 \"" +
					std::string{resp.event} + "\" with status \"" +
					std::string{resp.payload.get_string_or_empty("status")} +
					"\" and ref \"" + force_to_string(resp.ref) + "\"");
		}

		{
			ws.write(R"({"topic":"webs:markets","event":"phx_join","payload":{},"ref":"2"})");
			auto resp = to_bitcoinity_message(ws.read());
			if (resp.event != "phx_reply" or resp.payload.get_string_or_empty("status") != "ok" or not resp.ref or *resp.ref != "2")
				throw BitcoinityError("Server responded to message 2 \"" +
					std::string{resp.event} + "\" with status \"" +
					std::string{resp.payload.get_string_or_empty("status")} +
					"\" and ref \"" + force_to_string(resp.ref) + "\"");
		}

		{
			ws.write(
				R"({"topic":"webs:markets_)" + std::string{exchange} + "_" +
				std::string{currency} + R"(","event":"phx_join","payload":{},"ref":"3"})"
			);
			auto resp = to_bitcoinity_message(ws.read());
			if (resp.event != "phx_reply" or resp.payload.get_string_or_empty("status") != "ok" or not resp.ref or *resp.ref != "3")
				throw BitcoinityError("Server responded to message 3 \"" +
					std::string{resp.event} + "\" with status \"" +
					std::string{resp.payload.get_string_or_empty("status")} +
					"\" and ref \"" + force_to_string(resp.ref) + "\"");
		}
	}

	std::string join_keys (const std::vector<std::string>& keys) noexcept {
		try {
			std::string retval;
			for (const auto& key : keys) {
				retval += "/" + key;
			}

			return retval;
		}
		catch (...) {
			return {};
		}
	}
} //unnamed namespace

BitcoinityMissingKeyError::BitcoinityMissingKeyError(const std::vector<std::string>& keys, const std::string& json) :
	BitcoinityError("Missing key \"" + join_keys(keys)  + "\" from input json:\n" + json),
	m_keys(keys),
	m_json(json)
{ }

const std::vector<std::string>& BitcoinityMissingKeyError::keys() const noexcept {
	return m_keys;
}

const std::string& BitcoinityMissingKeyError::json() const noexcept {
	return m_json;
}

BitcoinityMessage::BitcoinityMessage (
	BitcoinityPayload&& payload_p,
	std::optional<std::string_view> ref_p,
	std::string_view event_p,
	std::string_view topic_p
) :
	payload(std::move(payload_p)),
	ref(std::move(ref_p)),
	event(std::move(event_p)),
	topic(std::move(topic_p))
{
}

BitcoinityReader::BitcoinityReader(const std::string& user_agent, std::string_view exchange, std::string_view currency) :
	WebsocketReader(std::string{BitcoinityHostname},
		std::string{BitcoinityPort},
		std::string{BitcoinityWSPath},
		user_agent
	)
{
	init_bitcoinity(*this, exchange, currency);
}

BitcoinityMessage BitcoinityReader::read() {
	return to_bitcoinity_message(WebsocketReader::read());
}
} //namespace duck
