/* Copyright 2022, Michele Santullo
 * This file is part of duckticker.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with duckticker.  If not, see <http://www.gnu.org/licenses/>.
 */

#define QUIET_SEND_RECV

#include "websocket_reader.hpp"
#include <boost/beast.hpp>
#include <boost/asio.hpp>
#if !defined(QUIET_SEND_RECV)
#	include <iostream>
#endif

namespace duck {
const std::string_view WebsocketReader::BeastVersionString = BOOST_BEAST_VERSION_STRING;

struct WebsocketReader::LocalData {
	using tcp = boost::asio::ip::tcp;

	LocalData() :
		resolver{ioc},
		ws{ioc}
	{ }

	boost::asio::io_context ioc;
	tcp::resolver resolver;
	boost::beast::websocket::stream<tcp::socket> ws;
};

WebsocketReader::WebsocketReader (std::string host,
		const std::string& port,
		const std::string& path,
		const std::string& user_agent
) :
	m_local(std::make_unique<LocalData>())
{
	//see https://www.boost.org/doc/libs/develop/libs/beast/example/websocket/client/sync/websocket_client_sync.cpp
	const auto results = m_local->resolver.resolve(host, port);
	auto ep = boost::asio::connect(m_local->ws.next_layer(), results);
	host += ':' + std::to_string(ep.port());
	m_local->ws.set_option(boost::beast::websocket::stream_base::decorator(
		[user_agent](boost::beast::websocket::request_type& req)
		{
			req.set(boost::beast::http::field::user_agent, user_agent);
		}
	));
	m_local->ws.handshake(host, path);
}

std::string WebsocketReader::read() {
	boost::beast::flat_buffer buffer;
	m_local->ws.read(buffer);
	std::string retval = boost::beast::buffers_to_string(buffer.data());
#if !defined(QUIET_SEND_RECV)
	std::cout << "RECV << " << retval << '\n';
#endif
	return retval;
}

void WebsocketReader::write(const std::string_view& message) {
#if !defined(QUIET_SEND_RECV)
	std::cout << "SEND >> " << message << '\n';
#endif
	m_local->ws.write(boost::asio::buffer(message));
}

WebsocketReader::~WebsocketReader() {
	m_local->ws.close(boost::beast::websocket::close_code::normal);
}
} //namespace duck
